@@ -1,6 +1,4 @@
 /*
- * $Id: afp_config.c,v 1.22.6.9 2004/06/14 20:58:56 bfernhomberg Exp $
- *
  * Copyright (c) 1997 Adrian Sun (asun@zoology.washington.edu)
  * All Rights Reserved.  See COPYRIGHT.
  */
@@ -13,33 +11,19 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
-
-/* STDC check */
-#if STDC_HEADERS
 #include <string.h>
-#else /* STDC_HEADERS */
-#ifndef HAVE_STRCHR
-#define strchr index
-#define strrchr index
-#endif /* HAVE_STRCHR */
-char *strchr (), *strrchr ();
-#ifndef HAVE_MEMCPY
-#define memcpy(d,s,n) bcopy ((s), (d), (n))
-#define memmove(d,s,n) bcopy ((s), (d), (n))
-#endif /* ! HAVE_MEMCPY */
-#endif /* STDC_HEADERS */
-
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif /* HAVE_UNISTD_H */
 #include <ctype.h>
-#include <atalk/logger.h>
-#include <atalk/util.h>
-
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#ifdef USE_SRVLOC
+#include <slp.h>
+#endif /* USE_SRVLOC */
+
+#include <atalk/logger.h>
+#include <atalk/util.h>
 #include <atalk/dsi.h>
 #include <atalk/atp.h>
 #include <atalk/asp.h>
@@ -47,14 +31,17 @@ char *strchr (), *strrchr ();
 #include <atalk/afp.h>
 #include <atalk/compat.h>
 #include <atalk/server_child.h>
-#ifdef USE_SRVLOC
-#include <slp.h>
-#endif /* USE_SRVLOC */
 
-#include "globals.h"
+#ifdef HAVE_LDAP
+#include <atalk/ldapconfig.h>
+#endif
+
+#include <atalk/globals.h>
 #include "afp_config.h"
 #include "uam_auth.h"
 #include "status.h"
+#include "volume.h"
+#include "afp_zeroconf.h"
 
 #define LINESIZE 1024  
 
@@ -93,10 +80,13 @@ void configfree(AFPConfig *configs, const AFPConfig *config)
         }
         free(p);
     }
+
+    /* the master loaded the volumes for zeroconf, get rid of that */
+    unload_volumes_and_extmap();
 }
 
 #ifdef USE_SRVLOC
-static void SRVLOC_callback(SLPHandle hslp, SLPError errcode, void *cookie) {
+static void SRVLOC_callback(SLPHandle hslp _U_, SLPError errcode, void *cookie) {
     *(SLPError*)cookie = errcode;
 }
 
@@ -114,7 +104,7 @@ static char * srvloc_encode(const struct afp_options *options, const char *name)
 
 	/* Convert name to maccharset */
         if ((size_t)-1 ==(convert_string_allocate( options->unixcharset, options->maccharset,
-			 name, strlen(name), &conv_name)) )
+			 name, -1, &conv_name)) )
 		return (char*)name;
 
 	/* Escape characters */
@@ -154,9 +144,9 @@ static char * srvloc_encode(const struct afp_options *options, const char *name)
 }
 #endif /* USE_SRVLOC */
 
-#ifdef USE_SRVLOC
 static void dsi_cleanup(const AFPConfig *config)
 {
+#ifdef USE_SRVLOC
     SLPError err;
     SLPError callbackerr;
     SLPHandle hslp;
@@ -189,8 +179,8 @@ static void dsi_cleanup(const AFPConfig *config)
 srvloc_dereg_err:
     dsi->srvloc_url[0] = '\0';
     SLPClose(hslp);
-}
 #endif /* USE_SRVLOC */
+}
 
 #ifndef NO_DDP
 static void asp_cleanup(const AFPConfig *config)
@@ -224,25 +214,22 @@ static int asp_start(AFPConfig *config, AFPConfig *configs,
 }
 #endif /* no afp/asp */
 
-static int dsi_start(AFPConfig *config, AFPConfig *configs,
-                     server_child *server_children)
+static afp_child_t *dsi_start(AFPConfig *config, AFPConfig *configs,
+                              server_child *server_children)
 {
-    DSI *dsi;
+    DSI *dsi = config->obj.handle;
+    afp_child_t *child = NULL;
 
-    if (!(dsi = dsi_getsession(config->obj.handle, server_children,
-                               config->obj.options.tickleval))) {
-        LOG(log_error, logtype_afpd, "main: dsi_getsession: %s", strerror(errno) );
-        exit( EXITERR_CLNT );
-    }
-
-    /* we've forked. */
-    if (dsi->child) {
+    if (!(child = dsi_getsession(dsi,
+                                 server_children,
+                                 config->obj.options.tickleval))) {
+        /* we've forked. */
         configfree(configs, config);
         afp_over_dsi(&config->obj); /* start a session */
         exit (0);
     }
 
-    return 0;
+    return child;
 }
 
 #ifndef NO_DDP
@@ -328,8 +315,8 @@ static AFPConfig *ASPConfigInit(const struct afp_options *options,
     config->optcount = refcount;
     (*refcount)++;
 
-    config->server_start = asp_start;
-    config->server_cleanup = asp_cleanup;
+    config->server_start = (void *) asp_start;
+    config->server_cleanup = (void *) asp_cleanup;
 
     return config;
 
@@ -348,20 +335,17 @@ static AFPConfig *DSIConfigInit(const struct afp_options *options,
     AFPConfig *config;
     DSI *dsi;
     char *p, *q;
-#ifdef USE_SRVLOC
-    SLPError err;
-    SLPError callbackerr;
-    SLPHandle hslp;
-    struct servent *afpovertcp;
-    int afp_port = 548;
-    char *srvloc_hostname, *hostname;
-#endif /* USE_SRVLOC */
 
     if ((config = (AFPConfig *) calloc(1, sizeof(AFPConfig))) == NULL) {
         LOG(log_error, logtype_afpd, "DSIConfigInit: malloc(config): %s", strerror(errno) );
         return NULL;
     }
 
+    LOG(log_debug, logtype_afpd, "DSIConfigInit: hostname: %s, ip/port: %s/%s, ",
+        options->hostname,
+        options->ipaddr ? options->ipaddr : "default",
+        options->port ? options->port : "548");
+
     if ((dsi = dsi_init(protocol, "afpd", options->hostname,
                         options->ipaddr, options->port,
                         options->flags & OPTION_PROXY,
@@ -370,20 +354,27 @@ static AFPConfig *DSIConfigInit(const struct afp_options *options,
         free(config);
         return NULL;
     }
+    dsi->dsireadbuf = options->dsireadbuf;
 
     if (options->flags & OPTION_PROXY) {
-        LOG(log_info, logtype_afpd, "ASIP proxy initialized for %s:%d (%s)",
-            inet_ntoa(dsi->server.sin_addr), ntohs(dsi->server.sin_port),
-            VERSION);
+        LOG(log_note, logtype_afpd, "AFP/TCP proxy initialized for %s:%d (%s)",
+            getip_string((struct sockaddr *)&dsi->server), getip_port((struct sockaddr *)&dsi->server), VERSION);
     } else {
-        LOG(log_info, logtype_afpd, "ASIP started on %s:%d(%d) (%s)",
-            inet_ntoa(dsi->server.sin_addr), ntohs(dsi->server.sin_port),
-            dsi->serversock, VERSION);
+        LOG(log_note, logtype_afpd, "AFP/TCP started, advertising %s:%d (%s)",
+            getip_string((struct sockaddr *)&dsi->server), getip_port((struct sockaddr *)&dsi->server), VERSION);
     }
 
 #ifdef USE_SRVLOC
     dsi->srvloc_url[0] = '\0';	/*  Mark that we haven't registered.  */
     if (!(options->flags & OPTION_NOSLP)) {
+        SLPError err;
+        SLPError callbackerr;
+        SLPHandle hslp;
+        unsigned int afp_port;
+        int   l;
+        char *srvloc_hostname;
+        const char *hostname;
+
 	err = SLPOpen("en", SLP_FALSE, &hslp);
 	if (err != SLP_OK) {
 	    LOG(log_error, logtype_afpd, "DSIConfigInit: Error opening SRVLOC handle");
@@ -396,10 +387,7 @@ static AFPConfig *DSIConfigInit(const struct afp_options *options,
 	 * use a non-default port, they can, but be aware, this server might
 	 * not show up int the Network Browser.
 	 */
-	afpovertcp = getservbyname("afpovertcp", "tcp");
-	if (afpovertcp != NULL) {
-	    afp_port = afpovertcp->s_port;
-	}
+	afp_port = getip_port((struct sockaddr *)&dsi->server);
 	/* If specified use the FQDN to register with srvloc, otherwise use IP. */
 	p = NULL;
 	if (options->fqdn) {
@@ -407,25 +395,27 @@ static AFPConfig *DSIConfigInit(const struct afp_options *options,
 	    p = strchr(hostname, ':');
 	}	
 	else 
-	    hostname = inet_ntoa(dsi->server.sin_addr);
+	    hostname = getip_string((struct sockaddr *)&dsi->server);
+
 	srvloc_hostname = srvloc_encode(options, (options->server ? options->server : options->hostname));
 
-	if (strlen(srvloc_hostname) > (sizeof(dsi->srvloc_url) - strlen(hostname) - 21)) {
+	if ((p) || afp_port == 548) {
+	    l = snprintf(dsi->srvloc_url, sizeof(dsi->srvloc_url), "afp://%s/?NAME=%s", hostname, srvloc_hostname);
+	}
+	else {
+	    l = snprintf(dsi->srvloc_url, sizeof(dsi->srvloc_url), "afp://%s:%d/?NAME=%s", hostname, afp_port, srvloc_hostname);
+	}
+
+	if (l == -1 || l >= (int)sizeof(dsi->srvloc_url)) {
 	    LOG(log_error, logtype_afpd, "DSIConfigInit: Hostname is too long for SRVLOC");
 	    dsi->srvloc_url[0] = '\0';
 	    goto srvloc_reg_err;
 	}
-	if ((p) || dsi->server.sin_port == afp_port) {
-	    sprintf(dsi->srvloc_url, "afp://%s/?NAME=%s", hostname, srvloc_hostname);
-	}
-	else {
-	    sprintf(dsi->srvloc_url, "afp://%s:%d/?NAME=%s", hostname, ntohs(dsi->server.sin_port), srvloc_hostname);
-	}
 
 	err = SLPReg(hslp,
 		     dsi->srvloc_url,
 		     SLP_LIFETIME_MAXIMUM,
-		     "",
+		     "afp",
 		     "",
 		     SLP_TRUE,
 		     SRVLOC_callback,
@@ -443,15 +433,16 @@ static AFPConfig *DSIConfigInit(const struct afp_options *options,
 	}
 
 	LOG(log_info, logtype_afpd, "Sucessfully registered %s with SRVLOC", dsi->srvloc_url);
+	config->server_cleanup = dsi_cleanup;
 
 srvloc_reg_err:
 	SLPClose(hslp);
     }
 #endif /* USE_SRVLOC */
 
-
     config->fd = dsi->serversock;
     config->obj.handle = dsi;
+    dsi->AFPobj = &config->obj;
     config->obj.config = config;
     config->obj.proto = AFPPROTO_DSI;
 
@@ -466,16 +457,13 @@ srvloc_reg_err:
     (*refcount)++;
 
     config->server_start = dsi_start;
-#ifdef USE_SRVLOC
-    config->server_cleanup = dsi_cleanup;
-#endif 
     return config;
 }
 
 /* allocate server configurations. this should really store the last
  * entry in config->last or something like that. that would make
  * supporting multiple dsi transports easier. */
-static AFPConfig *AFPConfigInit(const struct afp_options *options,
+static AFPConfig *AFPConfigInit(struct afp_options *options,
                                 const struct afp_options *defoptions)
 {
     AFPConfig *config = NULL, *next = NULL;
@@ -494,6 +482,10 @@ static AFPConfig *AFPConfigInit(const struct afp_options *options,
         config->defoptions = defoptions;
 #endif /* NO_DDP */
 
+
+    /* set signature */
+    set_signature(options);
+
     /* handle dsi transports and dsi proxies. we only proxy
      * for DSI connections. */
 
@@ -534,10 +526,10 @@ AFPConfig *configinit(struct afp_options *cmdline)
 {
     FILE *fp;
     char buf[LINESIZE + 1], *p, have_option = 0;
+    size_t len;
     struct afp_options options;
-    AFPConfig *config, *first = NULL;
+    AFPConfig *config=NULL, *first = NULL; 
 
-    status_reset();
     /* if config file doesn't exist, load defaults */
     if ((fp = fopen(cmdline->configfile, "r")) == NULL)
     {
@@ -546,12 +538,17 @@ AFPConfig *configinit(struct afp_options *cmdline)
         return AFPConfigInit(cmdline, cmdline);
     }
 
-    LOG(log_debug, logtype_afpd, "Loading ConfigFile"); 
-
     /* scan in the configuration file */
+    len = 0;
     while (!feof(fp)) {
-        if (!fgets(buf, sizeof(buf), fp) || buf[0] == '#')
+	if (!fgets(&buf[len], LINESIZE - len, fp) || buf[len] == '#')
             continue;
+	len = strlen(buf);
+	if ( len >= 2 && buf[len-2] == '\\' ) {
+	    len -= 2;
+	    continue;
+	} else
+	    len = 0;
 
         /* a little pre-processing to get rid of spaces and end-of-lines */
         p = buf;
@@ -566,7 +563,7 @@ AFPConfig *configinit(struct afp_options *cmdline)
         if (!afp_options_parseline(p, &options))
             continue;
 
-        /* this should really get a head and a tail to simplify things. */
+        /* AFPConfigInit can return two linked configs due to DSI and ASP */
         if (!first) {
             if ((first = AFPConfigInit(&options, cmdline)))
                 config = first->next ? first->next : first;
@@ -575,11 +572,22 @@ AFPConfig *configinit(struct afp_options *cmdline)
         }
     }
 
+#ifdef HAVE_LDAP
+    /* Parse afp_ldap.conf */
+    acl_ldap_readconfig(_PATH_ACL_LDAPCONF);
+#endif /* HAVE_LDAP */
+
     LOG(log_debug, logtype_afpd, "Finished parsing Config File");
     fclose(fp);
 
     if (!have_option)
-        return AFPConfigInit(cmdline, cmdline);
+        first = AFPConfigInit(cmdline, cmdline);
+
+    /* Now register with zeroconf, we also need the volumes for that */
+    if (first && !(first->obj.options.flags & OPTION_NOZEROCONF)) {
+        load_volumes(&first->obj);
+        zeroconf_register(first);
+    }
 
     return first;
 }
