@@ -1,6 +1,4 @@
 /*
- * $Id: unix.c,v 1.43.2.1.2.10 2004/06/15 22:53:55 didg Exp $
- *
  * Copyright (c) 1990,1993 Regents of The University of Michigan.
  * All Rights Reserved.  See COPYRIGHT.
  */
@@ -11,6 +9,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <inttypes.h>
 
 /* STDC check */
 #ifdef STDC_HEADERS
@@ -30,30 +29,31 @@ char *strchr (), *strrchr ();
 #endif /* STDC_HEADERS */
 
 #include <errno.h>
-#include <dirent.h>
 #include <limits.h>
 #include <sys/param.h>
-#include <sys/stat.h>
 #include <atalk/logger.h>
 #include <atalk/adouble.h>
+#include <atalk/vfs.h>
 #include <atalk/afp.h>
 #include <atalk/util.h>
+#include <atalk/unix.h>
+#include <atalk/acl.h>
 
 #include "auth.h"
 #include "directory.h"
 #include "volume.h"
 #include "unix.h"
 #include "fork.h"
+#ifdef HAVE_ACLS
+#include "acls.h"
+#endif
 
 /*
  * Get the free space on a partition.
  */
-int ustatfs_getvolspace( vol, bfree, btotal, bsize )
-const struct vol	*vol;
-VolSpace    *bfree, *btotal;
-u_int32_t   *bsize;
+int ustatfs_getvolspace(const struct vol *vol, VolSpace *bfree, VolSpace *btotal, u_int32_t *bsize)
 {
-    VolSpace maxVolSpace = (~(VolSpace)0);
+    VolSpace maxVolSpace = UINT64_MAX;
 
 #ifdef ultrix
     struct fs_data	sfs;
@@ -61,7 +61,6 @@ u_int32_t   *bsize;
     struct statfs	sfs;
 #endif /*ultrix*/
 
-
     if ( statfs( vol->v_path, &sfs ) < 0 ) {
         LOG(log_error, logtype_afpd, "ustatfs_getvolspace unable to stat %s", vol->v_path);
         return( AFPERR_PARAM );
@@ -85,8 +84,7 @@ u_int32_t   *bsize;
     *btotal = (VolSpace)
               ( sfs.fd_req.btot - ( sfs.fd_req.bfree - sfs.fd_req.bfreen ));
 #else /* !ultrix */
-    *btotal = (VolSpace)
-              ( sfs.f_blocks - ( sfs.f_bfree - sfs.f_bavail ));
+    *btotal = (VolSpace) sfs.f_blocks;
 #endif /* ultrix */
 
     /* see similar block above comments */
@@ -99,8 +97,7 @@ u_int32_t   *bsize;
     return( AFP_OK );
 }
 
-static __inline__ int utombits( bits )
-mode_t	bits;
+static int utombits(mode_t bits)
 {
     int		mbits;
 
@@ -117,9 +114,7 @@ mode_t	bits;
 /* --------------------------------
     cf AFP 3.0 page 63
 */
-void utommode( stat, ma )
-struct stat		*stat;
-struct maccess	*ma;
+void utommode(struct stat *stat, struct maccess *ma)
 {
 mode_t mode;
 
@@ -161,6 +156,7 @@ mode_t mode;
 }
 
 #ifdef accessmode
+
 #undef accessmode
 #endif
 /*
@@ -169,29 +165,26 @@ mode_t mode;
  *
  * Note: the previous method, using access(), does not work correctly
  * over NFS.
- * FIXME what about ACL?
+ *
+ * dir parameter is used by AFS
  */
-void accessmode( path, ma, dir, st )
-char		*path;
-struct maccess	*ma;
-struct dir	*dir;
-struct stat     *st;
-
+void accessmode(const struct vol *vol, char *path, struct maccess *ma, struct dir *dir _U_, struct stat *st)
 {
-struct stat     sb;
+    struct stat     sb;
 
     ma->ma_user = ma->ma_owner = ma->ma_world = ma->ma_group = 0;
     if (!st) {
-        if (stat(path, &sb) != 0)
+        if (ostat(path, &sb, vol_syml_opt(vol)) != 0)
             return;
         st = &sb;
     }
     utommode( st, ma );
-    return;
+#ifdef HAVE_ACLS
+    acltoownermode(vol, path, st, ma);
+#endif
 }
 
-int gmem( gid )
-const gid_t	gid;
+int gmem(const gid_t gid)
 {
     int		i;
 
@@ -203,8 +196,7 @@ const gid_t	gid;
     return( 0 );
 }
 
-static __inline__ mode_t mtoubits( bits )
-u_char	bits;
+static mode_t mtoubits(u_char bits)
 {
     mode_t	mode;
 
@@ -223,8 +215,7 @@ u_char	bits;
    and from AFP 3.0 spec page 63 
    the mac mode should be save somewhere 
 */
-mode_t mtoumode( ma )
-struct maccess	*ma;
+mode_t mtoumode(struct maccess *ma)
 {
     mode_t		mode;
 
@@ -240,80 +231,10 @@ struct maccess	*ma;
     return( mode );
 }
 
-/* ----------------------------- */
-char *fullpathname(const char *name)
-{
-    static char wd[ MAXPATHLEN + 1];
-
-    if ( getcwd( wd , MAXPATHLEN) ) {
-        strlcat(wd, "/", MAXPATHLEN);
-        strlcat(wd, name, MAXPATHLEN);
-    }
-    else {
-        strlcpy(wd, name, MAXPATHLEN);
-    }
-    return wd;
-}
-
-/* -----------------------------
-   a dropbox is a folder where w is set but not r eg:
-   rwx-wx-wx or rwx-wx-- 
-   rwx----wx (is not asked by a Mac with OS >= 8.0 ?)
-*/
-static int stickydirmode(name, mode, dropbox)
-char * name;
-const mode_t mode;
-const int dropbox;
-{
-    int retval = 0;
-
-#ifdef DROPKLUDGE
-    /* Turn on the sticky bit if this is a drop box, also turn off the setgid bit */
-    if (dropbox) {
-        int uid;
-
-        if ( ( (mode & S_IWOTH) && !(mode & S_IROTH)) ||
-             ( (mode & S_IWGRP) && !(mode & S_IRGRP)) )
-        {
-            uid=geteuid();
-            if ( seteuid(0) < 0) {
-                LOG(log_error, logtype_afpd, "stickydirmode: unable to seteuid root: %s", strerror(errno));
-            }
-            if ( (retval=chmod( name, ( (DIRBITS | mode | S_ISVTX) & ~default_options.umask) )) < 0) {
-                LOG(log_error, logtype_afpd, "stickydirmode: chmod \"%s\": %s", fullpathname(name), strerror(errno) );
-            } else {
-#ifdef DEBUG
-                LOG(log_info, logtype_afpd, "stickydirmode: (debug) chmod \"%s\": %s", fullpathname(name), strerror(retval) );
-#endif /* DEBUG */
-            }
-            seteuid(uid);
-            return retval;
-        }
-    }
-#endif /* DROPKLUDGE */
-
-    /*
-     *  Ignore EPERM errors:  We may be dealing with a directory that is
-     *  group writable, in which case chmod will fail.
-     */
-    if ( (chmod( name, (DIRBITS | mode) & ~default_options.umask ) < 0) && errno != EPERM)  {
-        LOG(log_error, logtype_afpd, "stickydirmode: chmod \"%s\": %s", fullpathname(name), strerror(errno) );
-        retval = -1;
-    }
-
-    return retval;
-}
-
-/* ------------------------- */
-int dir_rx_set(mode_t mode)
-{
-    return (mode & (S_IXUSR | S_IRUSR)) == (S_IXUSR | S_IRUSR);
-}
-
 #define EXEC_MODE (S_IXGRP | S_IXUSR | S_IXOTH)
 
-int setdeskmode( mode )
-const mode_t	mode;
+/* Using chmod_acl() instead of ochmod is ok here */
+int setdeskmode(const mode_t mode)
 {
     char		wd[ MAXPATHLEN + 1];
     struct stat         st;
@@ -362,17 +283,17 @@ const mode_t	mode;
             }
 
             if (S_ISDIR(st.st_mode)) {
-                if ( chmod( modbuf,  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
+                if ( chmod_acl( modbuf,  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
                      LOG(log_error, logtype_afpd, "setdeskmode: chmod %s: %s",fullpathname(modbuf), strerror(errno) );
                 }
-            } else if ( chmod( modbuf,  mode & ~(default_options.umask | EXEC_MODE) ) < 0 && errno != EPERM ) {
+            } else if ( chmod_acl( modbuf,  mode & ~(default_options.umask | EXEC_MODE) ) < 0 && errno != EPERM ) {
                 LOG(log_error, logtype_afpd, "setdeskmode: chmod %s: %s",fullpathname(modbuf), strerror(errno) );
             }
 
         }
         closedir( sub );
         /* XXX: need to preserve special modes */
-        if ( chmod( deskp->d_name,  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
+        if ( chmod_acl( deskp->d_name,  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
             LOG(log_error, logtype_afpd, "setdeskmode: chmod %s: %s",fullpathname(deskp->d_name), strerror(errno) );
         }
     }
@@ -382,115 +303,75 @@ const mode_t	mode;
         return -1;
     }
     /* XXX: need to preserve special modes */
-    if ( chmod( ".AppleDesktop",  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
+    if ( chmod_acl( ".AppleDesktop",  (DIRBITS | mode) & ~default_options.umask ) < 0 && errno != EPERM ) {
         LOG(log_error, logtype_afpd, "setdeskmode: chmod %s: %s", fullpathname(".AppleDesktop"),strerror(errno) );
     }
     return( 0 );
 }
 
 /* --------------------- */
-int setfilunixmode (vol, path, mode)
-const struct vol *vol;
-struct path* path;
-mode_t mode;
+int setfilunixmode (const struct vol *vol, struct path* path, mode_t mode)
 {
     if (!path->st_valid) {
-        of_stat(path);
+        of_stat(vol, path);
     }
 
     if (path->st_errno) {
         return -1;
     }
         
-    if (setfilmode( path->u_name, mode, &path->st) < 0)
+    mode |= vol->v_fperm;
+
+    if (setfilmode(vol, path->u_name, mode, &path->st) < 0)
         return -1;
     /* we need to set write perm if read set for resource fork */
-    return setfilmode(vol->ad_path( path->u_name, ADFLAGS_HF ), ad_hf_mode(mode), &path->st);
+    return vol->vfs->vfs_setfilmode(vol, path->u_name, mode, &path->st);
 }
 
-/* --------------------- */
-int setfilmode(name, mode, st)
-char * name;
-mode_t mode;
-struct stat *st;
-{
-struct stat sb;
-mode_t mask = S_IRUSR |S_IWUSR | S_IRGRP | S_IWGRP |S_IROTH | S_IWOTH;
-
-    if (!st) {
-        if (stat(name, &sb) != 0)
-            return -1;
-        st = &sb;
-    }
-   mode &= mask;	/* keep only rw-rw-rw in mode */
-   mode |= st->st_mode & ~mask; /* keep other bits from previous mode */
-   if ( chmod( name,  mode & ~default_options.umask ) < 0 && errno != EPERM ) {
-       return -1;
-   }
-   return 0;
-}
 
 /* --------------------- */
-int setdirunixmode( vol, name, mode )
-const struct vol *vol;
-const char       *name;
-const mode_t     mode;
+int setdirunixmode(const struct vol *vol, const char *name, mode_t mode)
 {
-char *adouble = vol->ad_path( name, ADFLAGS_DIR );
 
     int dropbox = (vol->v_flags & AFPVOL_DROPBOX);
 
+    LOG(log_debug, logtype_afpd, "setdirunixmode('%s', mode:%04o) {v_dperm:%04o}",
+        fullpathname(name), mode, vol->v_dperm);
+
+    mode |= vol->v_dperm;
+
     if (dir_rx_set(mode)) {
-    	/* extending right? dir first then .AppleDouble */
-    	if ( stickydirmode(name, DIRBITS | mode, dropbox) < 0 )
+    	/* extending right? dir first then .AppleDouble in rf_setdirmode */
+    	if ( stickydirmode(name, DIRBITS | mode, dropbox, vol->v_umask) < 0 )
         	return -1;
-    	if (vol->v_adouble != AD_VERSION2_OSX) {
-            if (stickydirmode(ad_dir(adouble), DIRBITS | mode, dropbox) < 0 && !vol_noadouble(vol)) {
-                return  -1 ;
-            }
-        }
     }
-    if (setfilmode(adouble, ad_hf_mode(mode), NULL) < 0 && !vol_noadouble(vol)) {
+    if (vol->vfs->vfs_setdirunixmode(vol, name, mode, NULL) < 0 && !vol_noadouble(vol)) {
         return  -1 ;
     }
     if (!dir_rx_set(mode)) {
-    	if (vol->v_adouble != AD_VERSION2_OSX) {
-            if (stickydirmode(ad_dir(adouble), DIRBITS | mode, dropbox) < 0 && !vol_noadouble(vol)) {
-                return  -1 ;
-            }
-        }
-    	if ( stickydirmode(name, DIRBITS | mode, dropbox) < 0 )
+    	if ( stickydirmode(name, DIRBITS | mode, dropbox, vol->v_umask) < 0 )
             return -1;
     }
     return 0;
 }
 
 /* --------------------- */
-int setdirmode( vol, name, mode )
-const struct vol *vol;
-const char       *name;
-const mode_t mode;
+int setdirmode(const struct vol *vol, const char *name, mode_t mode)
 {
-    char		buf[ MAXPATHLEN + 1];
     struct stat		st;
-    char		*m;
     struct dirent	*dirp;
     DIR			*dir;
+    mode_t              hf_mode;
     int                 osx = vol->v_adouble == AD_VERSION2_OSX;
-    int                 hf_mode = ad_hf_mode(mode);
     int                 dropbox = (vol->v_flags & AFPVOL_DROPBOX);
-    char                *adouble = vol->ad_path( name, ADFLAGS_DIR );
-    char                *adouble_p = ad_dir(adouble);
     
+    mode |= vol->v_dperm;
+    hf_mode = ad_hf_mode(mode);
+
     if (dir_rx_set(mode)) {
-    	/* extending right? dir first then .AppleDouble */
-    	if ( stickydirmode(name, DIRBITS | mode, dropbox) < 0 )
+    	/* extending right? dir first */
+    	if ( stickydirmode(name, DIRBITS | mode, dropbox, vol->v_umask) < 0 )
         	return -1;
-    	if (!osx) {
-            if (stickydirmode(adouble_p, DIRBITS | mode, dropbox) < 0 && !vol_noadouble(vol)) {
-                return  -1 ;
-            }
-        }
     }
     
     if (( dir = opendir( name )) == NULL ) {
@@ -503,7 +384,7 @@ const mode_t mode;
         if ( *dirp->d_name == '.' && (!osx || dirp->d_name[1] != '_')) {
             continue;
         }
-        if ( stat( dirp->d_name, &st ) < 0 ) {
+        if (ostat(dirp->d_name, &st, vol_syml_opt(vol)) < 0 ) {
             LOG(log_error, logtype_afpd, "setdirmode: stat %s: %s",dirp->d_name, strerror(errno) );
             continue;
         }
@@ -511,76 +392,28 @@ const mode_t mode;
         if (!S_ISDIR(st.st_mode)) {
            int setmode = (osx && *dirp->d_name == '.')?hf_mode:mode;
 
-           if (setfilmode(dirp->d_name, setmode, &st) < 0) {
+           if (setfilmode(vol, dirp->d_name, setmode, &st) < 0) {
+               closedir( dir );
                 LOG(log_error, logtype_afpd, "setdirmode: chmod %s: %s",dirp->d_name, strerror(errno) );
                 return -1;
            }
         }
-#if 0
-        /* Don't change subdir perm */
-        else if (S_ISDIR(st.st_mode)) {
-                if (stickydirmode(dirp->d_name, DIRBITS | mode, dropbox) < 0)
-                    return (-1);
-            } else if (stickydirmode(dirp->d_name, mode, dropbox) < 0)
-                return (-1);
-        }
-#endif
     }
     closedir( dir );
     
-    if (osx) {
-        goto setdirmode_noadouble;
+    if (vol->vfs->vfs_setdirmode(vol, name, mode, NULL) < 0 && !vol_noadouble(vol)) {
+        return  -1 ;
     }
-    
-    /* change perm of .AppleDouble's files
-    */
-    if (( dir = opendir( adouble_p )) == NULL ) {
-        if (vol_noadouble(vol))
-            goto setdirmode_noadouble;
-        LOG(log_error, logtype_afpd, "setdirmode: opendir %s: %s", fullpathname(".AppleDouble"),strerror(errno) );
-        return( -1 );
-    }
-    strcpy( buf, adouble_p);
-    strcat( buf, "/" );
-    m = strchr( buf, '\0' );
-    for ( dirp = readdir( dir ); dirp != NULL; dirp = readdir( dir )) {
-        if ( strcmp( dirp->d_name, "." ) == 0 ||
-                strcmp( dirp->d_name, ".." ) == 0 ) {
-            continue;
-        }
-        *m = '\0';
-        strcat( buf, dirp->d_name );
-
-        if ( stat( buf, &st ) < 0 ) {
-            LOG(log_error, logtype_afpd, "setdirmode: stat %s: %s", buf, strerror(errno) );
-            continue;
-        }
-        if (!S_ISDIR(st.st_mode)) {
-           if (setfilmode(buf, hf_mode , &st) < 0) {
-               /* FIXME what do we do then? */
-           }
-        }
-    } /* end for */
-    closedir( dir );
 
     if (!dir_rx_set(mode)) {
-        /* XXX: need to preserve special modes */
-        if (stickydirmode(adouble_p, DIRBITS | mode, dropbox) < 0 ) {
-                return  -1 ;
-        }
-    }
-
-setdirmode_noadouble:
-    if (!dir_rx_set(mode)) {
-    	if ( stickydirmode(name, DIRBITS | mode, dropbox) < 0 )
+    	if ( stickydirmode(name, DIRBITS | mode, dropbox, vol->v_umask) < 0 )
         	return -1;
     }
     return( 0 );
 }
 
-int setdeskowner( uid, gid )
-const uid_t	uid;
-const gid_t	gid;
+/* ----------------------------- */
+int setdeskowner(const uid_t uid, const gid_t gid)
 {
     char		wd[ MAXPATHLEN + 1];
     char		modbuf[12 + 1], *m;
@@ -642,65 +475,43 @@ const gid_t	gid;
 }
 
 /* ----------------------------- */
-int setfilowner(vol, uid, gid, path)
-const struct vol *vol;
-const uid_t	uid;
-const gid_t	gid;
-struct path* path;
+int setfilowner(const struct vol *vol, const uid_t uid, const gid_t gid, struct path* path)
 {
-    struct stat st;
-    char  *ad_p;
 
     if (!path->st_valid) {
-        of_stat(path);
+        of_stat(vol, path);
     }
 
     if (path->st_errno) {
         return -1;
     }
 
-    if ( chown( path->u_name, uid, gid ) < 0 && errno != EPERM ) {
+    if (ochown( path->u_name, uid, gid, vol_syml_opt(vol)) < 0 && errno != EPERM ) {
         LOG(log_debug, logtype_afpd, "setfilowner: chown %d/%d %s: %s",
             uid, gid, path->u_name, strerror(errno) );
 	return -1;
     }
 
-    ad_p = vol->ad_path( path->u_name, ADFLAGS_HF );
-
-    if ( stat( ad_p, &st ) < 0 ) {
-	/* ignore */
-        return 0;
-    }
-    if ( chown( ad_p, uid, gid ) < 0 &&
-            errno != EPERM ) {
-        LOG(log_debug, logtype_afpd, "setfilowner: chown %d/%d %s: %s",
-            uid, gid, ad_p, strerror(errno) );
+    if (vol->vfs->vfs_chown(vol, path->u_name, uid, gid ) < 0 && errno != EPERM) {
+        LOG(log_debug, logtype_afpd, "setfilowner: rf_chown %d/%d %s: %s",
+            uid, gid, path->u_name, strerror(errno) );
         return -1;
     }
+
     return 0;
 }
 
-
 /* --------------------------------- 
  * uid/gid == 0 need to be handled as special cases. they really mean
  * that user/group should inherit from other, but that doesn't fit
  * into the unix permission scheme. we can get around this by
  * co-opting some bits. */
-int setdirowner(vol, name, uid, gid )
-const struct vol *vol;
-const char      *name;
-const uid_t	uid;
-const gid_t	gid;
+int setdirowner(const struct vol *vol, const char *name, const uid_t uid, const gid_t gid)
 {
-    char		buf[ MAXPATHLEN + 1];
     struct stat		st;
-    char		*m;
     struct dirent	*dirp;
     DIR			*dir;
     int                 osx = vol->v_adouble == AD_VERSION2_OSX;
-    int                 noadouble = vol_noadouble(vol);
-    char                *adouble; 
-    char                *adouble_p;
 
     if (( dir = opendir( name )) == NULL ) {
         return( -1 );
@@ -709,13 +520,13 @@ const gid_t	gid;
         if ( *dirp->d_name == '.' && (!osx || dirp->d_name[1] != '_')) {
             continue;
         }
-        if ( stat( dirp->d_name, &st ) < 0 ) {
+        if (ostat(dirp->d_name, &st, vol_syml_opt(vol)) < 0 ) {
             LOG(log_error, logtype_afpd, "setdirowner: stat %s: %s",
                 fullpathname(dirp->d_name), strerror(errno) );
             continue;
         }
         if (( st.st_mode & S_IFMT ) == S_IFREG ) {
-            if ( chown( dirp->d_name, uid, gid ) < 0 && errno != EPERM ) {
+            if (ochown(dirp->d_name, uid, gid, vol_syml_opt(vol)) < 0 && errno != EPERM ) {
                 LOG(log_debug, logtype_afpd, "setdirowner: chown %s: %s",
                     fullpathname(dirp->d_name), strerror(errno) );
                 /* return ( -1 ); Sometimes this is okay */
@@ -723,56 +534,15 @@ const gid_t	gid;
         }
     }
     closedir( dir );
+
+    if (vol->vfs->vfs_setdirowner(vol, name, uid, gid) < 0) {
+        return -1;
+    }
     
-    if (osx) {
-       goto setdirowner_noadouble;
-    }
-
-    adouble = vol->ad_path( name, ADFLAGS_DIR );
-    adouble_p = ad_dir(adouble);
-    if (( dir = opendir( adouble_p )) == NULL ) {
-        if (noadouble)
-            goto setdirowner_noadouble;
+    if (ostat(".", &st, vol_syml_opt(vol)) < 0 ) {
         return( -1 );
     }
-    strcpy( buf, adouble_p );
-    strcat( buf, "/" );
-    m = strchr( buf, '\0' );
-    for ( dirp = readdir( dir ); dirp != NULL; dirp = readdir( dir )) {
-        if ( strcmp( dirp->d_name, "." ) == 0 ||
-                strcmp( dirp->d_name, ".." ) == 0 ) {
-            continue;
-        }
-        *m = '\0';
-        strcat( buf, dirp->d_name );
-        if ( chown( buf, uid, gid ) < 0 && errno != EPERM ) {
-            LOG(log_debug, logtype_afpd, "setdirowner: chown %d/%d %s: %s",
-                uid, gid, fullpathname(buf), strerror(errno) );
-            /* return ( -1 ); Sometimes this is okay */
-        }
-    }
-    closedir( dir );
-
-    /*
-     * We cheat: we know that chown doesn't do anything.
-     */
-    if ( stat( ".AppleDouble", &st ) < 0 ) {
-        LOG(log_error, logtype_afpd, "setdirowner: stat %s: %s", fullpathname(".AppleDouble"), strerror(errno) );
-        return( -1 );
-    }
-    if ( gid && gid != st.st_gid && chown( ".AppleDouble", uid, gid ) < 0 &&
-            errno != EPERM ) {
-        LOG(log_debug, logtype_afpd, "setdirowner: chown %d/%d %s: %s",
-            uid, gid,fullpathname(".AppleDouble"), strerror(errno) );
-        /* return ( -1 ); Sometimes this is okay */
-    }
-
-setdirowner_noadouble:
-    if ( stat( ".", &st ) < 0 ) {
-        return( -1 );
-    }
-    if ( gid && gid != st.st_gid && chown( ".", uid, gid ) < 0 &&
-            errno != EPERM ) {
+    if ( gid && gid != st.st_gid && ochown(".", uid, gid, vol_syml_opt(vol)) < 0 && errno != EPERM ) {
         LOG(log_debug, logtype_afpd, "setdirowner: chown %d/%d %s: %s",
             uid, gid, fullpathname("."), strerror(errno) );
     }
@@ -827,45 +597,3 @@ recursive_chown_end:
 }
 #endif
 
-/* This is equivalent of unix rename(). */
-int unix_rename(const char *oldpath, const char *newpath)
-{
-#if 0
-	char pd_name[PATH_MAX+1];
-	int i;
-        struct stat pd_stat;
-        uid_t uid;
-#endif
-
-	if (rename(oldpath, newpath) < 0)
-		return -1;
-#if 0
-	for (i = 0; i <= PATH_MAX && newpath[i] != '\0'; i++)
-		pd_name[i] = newpath[i];
-	pd_name[i] = '\0';
-
-	while (i > 0 && pd_name[i] != '/') i--;
-	if (pd_name[i] == '/') i++;
-
-        pd_name[i++] = '.'; pd_name[i++] = '\0';
-
-        if (stat(pd_name, &pd_stat) < 0) {
-	    LOG(log_error, logtype_afpd, "stat() of parent dir failed: pd_name = %s, uid = %d: %s",
-		pd_name, geteuid(), strerror(errno));
-		return 0;
-	}
-
-	/* So we have SGID bit set... */
-        if ((S_ISGID & pd_stat.st_mode)	!= 0) {
-            uid = geteuid();
-            if (seteuid(0) < 0)
-		LOG(log_error, logtype_afpd, "seteuid() failed: %s", strerror(errno));
-            if (recursive_chown(newpath, uid, pd_stat.st_gid) < 0)
-		LOG(log_error, logtype_afpd, "chown() of parent dir failed: newpath=%s, uid=%d: %s",
-		    pd_name, geteuid(), strerror(errno));
-            seteuid(uid);
-	}
-#endif
-	return 0;
-}
-
