@@ -201,12 +201,12 @@ static int check_symlink(const char *name, int *adflags)
 
     /* fchdir() to pathbuf so we can easily get its path with getcwd() */
     if ((chdir(pathbuf)) != 0) {
-        dbd_log(LOGSTD, "Cant chdir to '%s': %s", pathbuf, strerror(errno));
+        dbd_log(LOGSTD, "Can't chdir to '%s': %s", pathbuf, strerror(errno));
         return -1;
     }
 
     if ((getcwd(pathbuf, MAXPATHLEN)) == NULL) {
-        dbd_log(LOGSTD, "Cant get symlink'ed dir '%s/%s': %s", cwdbuf, pathbuf, strerror(errno));
+        dbd_log(LOGSTD, "Can't get symlink'ed dir '%s/%s': %s", cwdbuf, pathbuf, strerror(errno));
         if ((fchdir(cwd)) != 0)
             /* We're foobared */
             longjmp(jmp, 1); /* this jumps back to cmd_dbd_scanvol() */
@@ -316,7 +316,7 @@ static int check_adfile(const char *fname, const struct stat *st)
         dbd_log(LOGSTD, "Missing AppleDouble file '%s/%s'", cwdbuf, adname);
 
         if (dbd_flags & DBD_FLAGS_SCAN)
-            /* Scan only requested, dont change anything */
+            /* Scan only requested, don't change anything */
             return -1;
 
         /* Create ad file */
@@ -374,7 +374,7 @@ static void remove_eafiles(const char *name, struct ea *ea)
     if ((dp = opendir(".")) == NULL) {
         dbd_log(LOGSTD, "Couldn't open the directory '%s/%s': %s",
                 cwdbuf, ADv2_DIRNAME, strerror(errno));
-        return;
+        goto exit;
     }
 
     while ((ep = readdir(dp))) {
@@ -388,9 +388,14 @@ static void remove_eafiles(const char *name, struct ea *ea)
         } /* if */
     } /* while */
 
+exit:
     if (dp)
         closedir(dp);
-
+    if ((chdir("..")) != 0) {
+        dbd_log(LOGSTD, "Couldn't chdir to '%s': %s", cwdbuf, strerror(errno));
+        /* we can't proceed */
+        longjmp(jmp, 1); /* this jumps back to cmd_dbd_scanvol() */
+    }    
 }
 
 /*
@@ -484,7 +489,7 @@ static int check_addir(int volroot)
         /* Yes, but are we only scanning ? */
         if (dbd_flags & DBD_FLAGS_SCAN) {
             /* Yes:  missing .Parent is not a problem, but missing ad-dir
-               causes later checking of ad-files to fail. So we have to return appropiately */
+               causes later checking of ad-files to fail. So we have to return appropriately */
             if (addir_ok != 0)
                 return -1;
             else  /* (adpar_ok != 0) */
@@ -561,7 +566,7 @@ static int check_eafile_in_adouble(const char *name)
                         cwdbuf, ADv2_DIRNAME, name);
 
                 if (dbd_flags & DBD_FLAGS_SCAN)
-                    /* Scan only requested, dont change anything */
+                    /* Scan only requested, don't change anything */
                     goto ea_check_done;
 
                 if ((unlink(name)) != 0) {
@@ -640,7 +645,7 @@ static int read_addir(void)
                     cwdbuf, ADv2_DIRNAME, ep->d_name);
 
             if (dbd_flags & DBD_FLAGS_SCAN)
-                /* Scan only requested, dont change anything */
+                /* Scan only requested, don't change anything */
                 continue;;
 
             if ((unlink(ep->d_name)) != 0) {
@@ -706,7 +711,7 @@ static cnid_t check_cnid(const char *name, cnid_t did, struct stat *st, int adfi
             ad_flush(&ad);
         }
         else
-            ad_cnid = ad_getid(&ad, st->st_dev, st->st_ino, did, stamp);
+            ad_cnid = ad_getid(&ad, st->st_dev, st->st_ino, 0, stamp);
 
         if (ad_cnid == 0)
             dbd_log( LOGSTD, "Bad CNID in adouble file of '%s/%s'", cwdbuf, name);
@@ -750,24 +755,22 @@ static cnid_t check_cnid(const char *name, cnid_t did, struct stat *st, int adfi
         /* Everything is fine */
         return db_cnid;
     } else if (ad_cnid && db_cnid && (ad_cnid != db_cnid)) {
-        /* Mismatch ? Delete both from db and re-add data from file */
+        /* Mismatch, overwrite ad file with value from db */
         dbd_log( LOGSTD, "CNID mismatch for '%s/%s', db: %u, ad-file: %u", cwdbuf, name, ntohl(db_cnid), ntohl(ad_cnid));
         if ( ! (dbd_flags & DBD_FLAGS_SCAN)) {
-            rqst.cnid = db_cnid;
-            ret = dbd_delete(dbd, &rqst, &rply, DBIF_CNID);
-            if (dbif_txn_close(dbd, ret) != 0)
-                return CNID_INVALID;
-
-            rqst.cnid = ad_cnid;
-            ret = dbd_delete(dbd, &rqst, &rply, DBIF_CNID);
-            if (dbif_txn_close(dbd, ret) != 0)
-                return CNID_INVALID;
-
-            ret = dbd_rebuild_add(dbd, &rqst, &rply);
-            if (dbif_txn_close(dbd, ret) != 0)
+            dbd_log(LOGSTD, "Updating AppleDouble file for '%s/%s' with CNID: %u from database",
+                            cwdbuf, name, ntohl(db_cnid));
+            ad_init(&ad, myvolinfo->v_adouble, myvolinfo->v_ad_options);
+            if (ad_open_metadata( name, adflags, O_RDWR, &ad) != 0) {
+                dbd_log(LOGSTD, "Error opening AppleDouble file for '%s/%s': %s",
+                        cwdbuf, name, strerror(errno));
                 return CNID_INVALID;
+            }
+            ad_setid( &ad, st->st_dev, st->st_ino, db_cnid, did, stamp);
+            ad_flush(&ad);
+            ad_close_metadata(&ad);
         }
-        return ad_cnid;
+        return db_cnid;
     } else if (ad_cnid && (db_cnid == 0)) {
         /* in ad-file but not in db */
         if ( ! (dbd_flags & DBD_FLAGS_SCAN)) {
@@ -777,7 +780,7 @@ static cnid_t check_cnid(const char *name, cnid_t did, struct stat *st, int adfi
 
             rqst.cnid = ad_cnid;
             ret = dbd_resolve(dbd, &rqst, &rply);
-            if (ret == CNID_DBD_RES_OK) {
+            if (rply.result == CNID_DBD_RES_OK) {
                 /* Occupied! Choose another, update ad-file */
                 ret = dbd_add(dbd, &rqst, &rply, 1);
                 if (dbif_txn_close(dbd, ret) != 0)
@@ -1012,11 +1015,11 @@ static int dbd_readdir(int volroot, cnid_t did)
             strcat(cwdbuf, ep->d_name);
             dbd_log( LOGDEBUG, "Entering directory: %s", cwdbuf);
             if (-1 == (cwd = open(".", O_RDONLY))) {
-                dbd_log( LOGSTD, "Cant open directory '%s': %s", cwdbuf, strerror(errno));
+                dbd_log( LOGSTD, "Can't open directory '%s': %s", cwdbuf, strerror(errno));
                 continue;
             }
             if (0 != chdir(ep->d_name)) {
-                dbd_log( LOGSTD, "Cant chdir to directory '%s': %s", cwdbuf, strerror(errno));
+                dbd_log( LOGSTD, "Can't chdir to directory '%s': %s", cwdbuf, strerror(errno));
                 close(cwd);
                 continue;
             }
@@ -1201,10 +1204,10 @@ int cmd_dbd_scanvol(DBD *dbd_ref, struct volinfo *vi, dbd_flags_t flags)
     /* Make it accessible for all funcs */
     dbd = dbd_ref;
 
-    /* We only support unicode volumes ! */
+    /* Officially, we only support unicode volumes */
     if ( vi->v_volcharset != CH_UTF8) {
         dbd_log( LOGSTD, "Not a Unicode volume: %s, %u != %u", vi->v_volcodepage, vi->v_volcharset, CH_UTF8);
-        return -1;
+        /* Log the unexpected charset and proceed */
     }
 
     /* Get volume stamp */
