@@ -1,5 +1,5 @@
 /*
- * $Id: file.h,v 1.6 2001/06/25 20:13:45 rufustfirefly Exp $
+ * $Id: file.h,v 1.8 2009-10-13 22:55:37 didg Exp $
  *
  * Copyright (c) 1990,1991 Regents of The University of Michigan.
  * All Rights Reserved.  See COPYRIGHT.
@@ -17,11 +17,15 @@ struct papfile {
     int			pf_datalen;
     char		*pf_buf;
     char		*pf_data;
+    int		origin;
 };
 
 #define PF_BOT		(1<<0)
 #define PF_EOF		(1<<1)
 #define PF_QUERY	(1<<2)
+#define PF_STW		(1<<3)
+#define PF_TRANSLATE	(1<<4)
+#define PF_FONT_QUERY	(1<<5)
 
 #define CONSUME( pf, len )  {   (pf)->pf_data += (len); \
 				(pf)->pf_datalen -= (len); \
@@ -33,9 +37,10 @@ struct papfile {
 
 #define PF_MORESPACE	1024
 
-int markline __P(( struct papfile *, char **, int *, int * ));
-void morespace __P(( struct papfile *, const char *, int ));
-void append __P(( struct papfile *, const char *, int ));
-void spoolerror __P(( struct papfile *, char * ));
+int markline ( struct papfile *, char **, int *, int * );
+void morespace ( struct papfile *, const char *, int );
+void append ( struct papfile *, const char *, int );
+void spoolerror ( struct papfile *, char * );
+void spoolreply ( struct papfile *, char * );
 
 #endif /* PAPD_FILE_H */
