@@ -1,5 +1,5 @@
 /*
- * $Id: print_cups.c,v 1.1.2.1.2.1 2005/01/11 23:00:40 didg Exp $
+ * $Id: print_cups.c,v 1.6 2010-01-26 20:43:11 didg Exp $
  *
  * Copyright 2004 Bjoern Fernhomberg.
  *
@@ -39,6 +39,7 @@
 
 #ifdef HAVE_CUPS
 
+#include <cups/ipp.h>
 #include <cups/cups.h>
 #include <cups/language.h>
 #include <atalk/unicode.h>
@@ -60,13 +61,13 @@ static const char* cups_status_msg[] = {
 };
 
 /* Local functions */
-static int 	convert_to_mac_name ( char *encoding, char * inptr, char * outptr, size_t outlen);
+static int 	convert_to_mac_name ( const char *encoding, char * inptr, char * outptr, size_t outlen);
 static size_t	to_ascii ( char *inbuf, char **outbuf);
 static int 	cups_mangle_printer_name ( struct printer *pr, struct printer *printers);
 static void     cups_free_printer ( struct printer *pr);
 
 
-char * cups_get_language ()
+const char * cups_get_language (void)
 {
         cups_lang_t *language;
 
@@ -79,7 +80,7 @@ char * cups_get_language ()
  */
 
 static const char *                            /* O - Password or NULL */
-cups_passwd_cb(const char *prompt)      /* I - Prompt */
+cups_passwd_cb(const char *prompt _U_)      /* I - Prompt */
 {
  /*
   * Always return NULL to indicate that no password is available...
@@ -96,6 +97,7 @@ int                                     /* O - 1 if printer name OK */
 cups_printername_ok(char *name)         /* I - Name of printer */
 {
         http_t          *http;          /* HTTP connection to server */
+	cups_dest_t	*dest = NULL;	/* Destination */
         ipp_t           *request,       /* IPP Request */
                         *response;      /* IPP Response */
         cups_lang_t     *language;      /* Default language */
@@ -107,17 +109,25 @@ cups_printername_ok(char *name)         /* I - Name of printer */
 
         cupsSetPasswordCB(cups_passwd_cb);
 
-       /*
-        * Try to connect to the server...
-        */
-
-        if ((http = httpConnect(cupsServer(), ippPort())) == NULL)
-        {
-		LOG(log_error, logtype_papd, "Unable to connect to CUPS server %s - %s",
-                         cupsServer(), strerror(errno));
-                return (0);
-        }
+	/*
+	 * Try to connect to the requested printer...
+	 */
 
+	dest = cupsGetNamedDest(CUPS_HTTP_DEFAULT, name, NULL);
+	
+	if (!dest)
+	{
+		LOG(log_error, logtype_papd,
+		    "Unable to get destination \"%s\": %s", name, cupsLastErrorString());
+		return (0);
+	}
+	if ((http = cupsConnectDest(dest, CUPS_DEST_FLAGS_NONE, 30000, NULL, NULL, 0, NULL, NULL)) == NULL)
+	{
+		LOG(log_error, logtype_papd,
+		    "Unable to connect to destination \"%s\": %s", dest->name, cupsLastErrorString());
+		return (0);
+	}
+	cupsFreeDests(1, dest);
 
        /*
         * Build an IPP_GET_PRINTER_ATTRS request, which requires the following
@@ -128,11 +138,7 @@ cups_printername_ok(char *name)         /* I - Name of printer */
         *    requested-attributes
         *    printer-uri
         */
-
-        request = ippNew();
-
-        request->request.op.operation_id = IPP_GET_PRINTER_ATTRIBUTES;
-        request->request.op.request_id   = 1;
+        request = ippNewRequest(IPP_OP_GET_PRINTER_ATTRIBUTES);
 
         language = cupsLangDefault();
 
@@ -164,10 +170,10 @@ cups_printername_ok(char *name)         /* I - Name of printer */
 
         httpClose(http);
 
-        if (response->request.status.status_code >= IPP_OK_CONFLICT)
+        if (cupsLastError() >= IPP_OK_CONFLICT)
         {
       		LOG(log_error, logtype_papd, "Unable to get printer status for %s - %s", name,
-                         ippErrorString(response->request.status.status_code));
+                         ippErrorString(cupsLastError()));
                 ippDelete(response);
                 return (0);
         }
@@ -183,8 +189,205 @@ cups_printername_ok(char *name)         /* I - Name of printer */
 const char * 
 cups_get_printer_ppd ( char * name)
 {
+	http_t		*http;		/* Connection to destination */
+	cups_dest_t	*dest = NULL;	/* Destination */
+	cups_dest_t 	*dests;		/* Destination List */
+	ipp_t           *request,       /* IPP Request */
+			*response;      /* IPP Response */
+	cups_lang_t	*language;      /* Default language */
+	char		uri[HTTP_MAX_URI]; /* printer-uri attribute */
+        const char	*pattrs[] =   /* Requested printer attributes */
+                        {
+                          "printer-make-and-model",
+                          "color-supported"
+                        };
+	ipp_attribute_t	*attr;
+  	cups_file_t	*fp;		/* PPD file */
+    	static char	buffer[1024];	/* I - Filename buffer */
+    	size_t		bufsize;	/* I - Size of filename buffer */
+  	char		make[256],	/* Make and model */
+			*model;
+
+
 	cupsSetPasswordCB(cups_passwd_cb);
-	return cupsGetPPD( name );
+
+	/*
+	 *We have to go this roundabout way to correctly get the make and 
+	 *model of DNS-SD discovered printers. Otherwise we get the name of 
+	 *the CUPS temporary queue, which we don't want.
+	 */
+
+	int num_dests = cupsGetDests2(CUPS_HTTP_DEFAULT, &dests);
+	dest = cupsGetDest(name, NULL, num_dests, dests);
+	const char *make_model = cupsGetOption("printer-make-and-model", dest->num_options, dest->options);
+
+	if (!dest)
+	{
+		LOG(log_error, logtype_papd, "testdest: Unable to get destination \"%s\": %s\n", name, cupsLastErrorString());
+		cupsFreeDests(num_dests,dests);
+		return (0);
+	}
+
+	if ((http = cupsConnectDest(dest, CUPS_DEST_FLAGS_NONE, 30000, NULL, NULL, 0, NULL, NULL)) == NULL)
+	{
+		LOG(log_error, logtype_papd, "testdest: Unable to connect to destination \"%s\": %s\n", dest->name, cupsLastErrorString());
+		cupsFreeDests(num_dests,dests);
+		return (0);
+	}
+	
+	/*
+	 * Generate the printer URI...
+	 */
+
+	sprintf(uri, "ipp://localhost/printers/%s", name);
+
+	/*
+	 * Build an IPP_GET_PRINTER_ATTRIBUTES request, which requires the
+	 * following attributes:
+	 *
+	 *    attributes-charset
+	 *    attributes-natural-language
+	 *    requested-attributes
+	 *    printer-uri
+	 */
+
+        request = ippNew();
+
+        ippSetOperation(request, IPP_GET_PRINTER_ATTRIBUTES);
+        ippSetRequestId(request, 1);
+
+	language = cupsLangDefault();
+
+	ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
+		     "attributes-charset", NULL,
+		     cupsLangEncoding(language));
+
+	ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
+		     "attributes-natural-language", NULL,
+		     language->language);
+
+	ippAddStrings(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+		      "requested-attributes",
+		      (sizeof(pattrs) / sizeof(pattrs[0])), NULL, pattrs);
+
+	ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
+		     "printer-uri", NULL, uri);
+
+	/*
+	 * Do the request and get back a response...
+	 */
+
+        if ((response = cupsDoRequest(http, request, "/")) == NULL)
+        {
+                 LOG(log_error, logtype_papd,  "Unable to get printer attribs for %s - %s", name,
+                         ippErrorString(cupsLastError()));
+                httpClose(http);
+		cupsFreeDests(num_dests,dests);
+                return (0);
+        }
+
+        if (ippGetStatusCode(response) >= IPP_OK_CONFLICT)
+        {
+      		 LOG(log_error, logtype_papd,  "Unable to get printer attribs for %s - %s", name,
+                         ippErrorString(ippGetStatusCode(response)));
+                ippDelete(response);
+                httpClose(http);
+		cupsFreeDests(num_dests,dests);
+                return (0);
+        }
+
+	/*
+	 * Range check input...
+	 */
+	
+	bufsize = sizeof(buffer);
+	if (buffer)
+		*buffer = '\0';
+
+	if (!buffer || bufsize < 1)
+	{
+		LOG(log_error, logtype_papd, "Check buffer failed!\n");
+		LOG(log_error, logtype_papd, strerror(EINVAL));
+                ippDelete(response);
+                httpClose(http);
+		cupsFreeDests(num_dests,dests);
+		return (0);
+	}
+
+	if (!response)
+	{
+		LOG(log_error, logtype_papd, "No IPP attributes.\n");
+                ippDelete(response);
+                httpClose(http);
+		cupsFreeDests(num_dests,dests);
+		return (0);
+	}
+
+	/*
+	 * Open a temporary file for the PPD...
+	 */
+
+	if ((fp = cupsTempFile2(buffer, (int)bufsize)) == NULL)
+	{
+		LOG(log_error, logtype_papd, strerror(errno));
+                ippDelete(response);
+                httpClose(http);
+		cupsFreeDests(num_dests,dests);
+		return (0);
+	}
+
+	/*
+	 * Write the header and some hard coded defaults for the PPD file...
+	 */
+
+	cupsFilePuts(fp, "*PPD-Adobe: \"4.3\"\n");
+	cupsFilePuts(fp, "*FormatVersion: \"4.3\"\n");
+	cupsFilePrintf(fp, "*FileVersion: \"%d.%d\"\n", CUPS_VERSION_MAJOR, CUPS_VERSION_MINOR);
+	cupsFilePuts(fp, "*LanguageVersion: English\n");
+	cupsFilePuts(fp, "*LanguageEncoding: ISOLatin1\n");
+	cupsFilePuts(fp, "*PSVersion: \"(3010.000) 0\"\n");
+	cupsFilePuts(fp, "*LanguageLevel: \"3\"\n");
+	cupsFilePuts(fp, "*FileSystem: False\n");
+	cupsFilePuts(fp, "*PCFileName: \"ippeve.ppd\"\n");
+
+	if ((dests != NULL))
+		strcpy(make, make_model);
+	else
+		strcpy(make, "Unknown Printer");
+
+	if (!strncasecmp(make, "Hewlett Packard ", 16) || !strncasecmp(make, "Hewlett-Packard ", 16))
+	{
+		model = make + 16;
+		strcpy(make, "HP");
+  	}
+	else if ((model = strchr(make, ' ')) != NULL)
+		*model++ = '\0';
+  	else
+		model = "Unknown";
+
+	cupsFilePrintf(fp, "*Manufacturer: \"%s\"\n", make);
+	cupsFilePrintf(fp, "*ModelName: \"%s\"\n", model);
+	cupsFilePrintf(fp, "*Product: \"(%s %s)\"\n", make, model);
+	cupsFilePrintf(fp, "*NickName: \"%s %s\"\n", make, model);
+	cupsFilePrintf(fp, "*ShortNickName: \"%s %s\"\n", make, model);
+
+ 	if ((attr = ippFindAttribute(response, "color-supported", IPP_TAG_BOOLEAN)) != NULL && ippGetBoolean(attr, 0))
+  		cupsFilePuts(fp, "*ColorDevice: True\n");
+	else
+		cupsFilePuts(fp, "*ColorDevice: False\n");
+	cupsFilePuts(fp, "*TTRasterizer: Type42\n");
+	cupsFilePuts(fp, "*Resolution: 600dpi\n");
+	cupsFilePuts(fp, "*FaxSupport: None\n");
+
+	/*
+	 *Clean up.
+	 */
+	
+	httpClose(http);
+	ippDelete(response);
+	cupsFileClose(fp);
+	cupsFreeDests(num_dests,dests);
+	return (buffer);
 }
 
 int
@@ -192,6 +395,7 @@ cups_get_printer_status (struct printer *pr)
 {
 
         http_t          *http;          /* HTTP connection to server */
+	cups_dest_t	*dest = NULL;	/* Destination */
         ipp_t           *request,       /* IPP Request */
                         *response;      /* IPP Response */
         ipp_attribute_t *attr;          /* Current attribute */
@@ -212,16 +416,25 @@ cups_get_printer_status (struct printer *pr)
 
         cupsSetPasswordCB(cups_passwd_cb);
 
-       /*
-        * Try to connect to the server...
-        */
+	/*
+	 * Try to connect to the requested printer...
+	 */
 
-        if ((http = httpConnect(cupsServer(), ippPort())) == NULL)
-        {
-		LOG(log_error, logtype_papd, "Unable to connect to CUPS server %s - %s",
-                         cupsServer(), strerror(errno));
-                return (0);
-        }
+	dest = cupsGetNamedDest(CUPS_HTTP_DEFAULT, pr->p_printer, NULL);
+	
+	if (!dest)
+	{
+		LOG(log_error, logtype_papd,
+		    "Unable to get destination \"%s\": %s", pr->p_printer, cupsLastErrorString());
+		return (0);
+	}
+	if ((http = cupsConnectDest(dest, CUPS_DEST_FLAGS_NONE, 30000, NULL, NULL, 0, NULL, NULL)) == NULL)
+	{
+		LOG(log_error, logtype_papd,
+		    "Unable to connect to destination \"%s\": %s", dest->name, cupsLastErrorString());
+		return (0);
+	}
+	cupsFreeDests(1, dest);
 
        /*
         * Generate the printer URI...
@@ -241,10 +454,7 @@ cups_get_printer_status (struct printer *pr)
         *    printer-uri
         */
 
-        request = ippNew();
-
-        request->request.op.operation_id = IPP_GET_PRINTER_ATTRIBUTES;
-        request->request.op.request_id   = 1;
+        request = ippNewRequest(IPP_OP_GET_PRINTER_ATTRIBUTES);
 
         language = cupsLangDefault();
 
@@ -274,10 +484,10 @@ cups_get_printer_status (struct printer *pr)
                 return (0);
         }
 
-        if (response->request.status.status_code >= IPP_OK_CONFLICT)
+        if (cupsLastError() >= IPP_OK_CONFLICT)
         {
       		LOG(log_error, logtype_papd, "Unable to get printer status for %s - %s", pr->p_printer,
-                         ippErrorString(response->request.status.status_code));
+                         ippErrorString(cupsLastError()));
                 ippDelete(response);
                 httpClose(http);
                 return (0);
@@ -291,9 +501,9 @@ cups_get_printer_status (struct printer *pr)
 
         if ((attr = ippFindAttribute(response, "printer-state", IPP_TAG_ENUM)) != NULL)
         {
-                if (attr->values[0].integer == IPP_PRINTER_STOPPED)
+                if (ippGetInteger(attr, 0) == IPP_PRINTER_STOPPED)
 			status = 1;
-                else if (attr->values[0].integer == IPP_NOT_ACCEPTING)
+                else if (ippGetInteger(attr,0) == IPP_NOT_ACCEPTING)
 			status = 0;
 		else
 			status = 2;
@@ -301,14 +511,14 @@ cups_get_printer_status (struct printer *pr)
 
 	if ((attr = ippFindAttribute(response, "printer-is-accepting-jobs", IPP_TAG_BOOLEAN)) != NULL)
 	{
-		if ( attr->values[0].integer == 0 ) 
+		if ( ippGetBoolean(attr, 0) == 0 )
 			status = 0;
 	}
 		
 	snprintf ( pr->p_status, 255, cups_status_msg[status], pr->p_printer );
 
         if ((attr = ippFindAttribute(response, "printer-state-message", IPP_TAG_TEXT)) != NULL)
-		strncat ( pr->p_status, attr->values[0].string.text, 255-strlen(pr->p_status));
+		strncat ( pr->p_status, ippGetString(attr, 0, NULL), 255-strlen(pr->p_status));
 
         ippDelete(response);
 
@@ -384,7 +594,7 @@ cups_autoadd_printers ( struct printer	*defprinter, struct printer *printers)
         for  (i=0; i< num_dests; i++)
         {
 		if (( pr = (struct printer *)malloc( sizeof( struct printer )))	== NULL ) {
-			LOG(log_error, logtype_papd, "malloc: %m" );
+			LOG(log_error, logtype_papd, "malloc: %s", strerror(errno) );
 			exit( 1 );
 		}
 	
@@ -392,7 +602,7 @@ cups_autoadd_printers ( struct printer	*defprinter, struct printer *printers)
 
 		/* convert from CUPS to local encoding */
                 convert_string_allocate( add_charset(cupsLangEncoding(language)), CH_UNIX, 
-                                         dests[i].name, strlen(dests[i].name), &pr->p_u_name);
+                                         dests[i].name, -1, &pr->p_u_name);
 
 		/* convert CUPS name to Mac charset */
 		if ( convert_to_mac_name ( cupsLangEncoding(language), dests[i].name, name, sizeof(name)) <= 0)
@@ -403,7 +613,7 @@ cups_autoadd_printers ( struct printer	*defprinter, struct printer *printers)
 		}
 
 		if (( pr->p_name = (char *)malloc( strlen( name ) + 1 )) == NULL ) {
-			LOG(log_error, logtype_papd, "malloc: %m" );
+			LOG(log_error, logtype_papd, "malloc: %s", strerror(errno) );
 			exit( 1 );
 		}
 		strcpy( pr->p_name, name );
@@ -415,14 +625,14 @@ cups_autoadd_printers ( struct printer	*defprinter, struct printer *printers)
 		pr->p_flags |= P_CUPS_AUTOADDED;
 			
 		if (( pr->p_printer = (char *)malloc( strlen( dests[i].name ) + 1 )) == NULL ) {
-			LOG(log_error, logtype_papd, "malloc: %m" );
+			LOG(log_error, logtype_papd, "malloc: %s", strerror(errno) );
                		exit( 1 );
         	}
         	strcpy( pr->p_printer, dests[i].name );			
 
         	if ( (p = (char *) cups_get_printer_ppd ( pr->p_printer )) != NULL ) {
         		if (( pr->p_ppdfile = (char *)malloc( strlen( p ) + 1 )) == NULL ) {
-				LOG(log_error, logtype_papd, "malloc: %m" );
+				LOG(log_error, logtype_papd, "malloc: %s", strerror(errno) );
                			exit( 1 );
         		}
         		strcpy( pr->p_ppdfile, p );
@@ -493,7 +703,7 @@ to_ascii ( char  *inptr, char **outptr)
 	char *out, *osav;
 
 	if ( NULL == (out = (char*) malloc ( strlen ( inptr) + 1 )) ) {
-		LOG(log_error, logtype_papd, "malloc: %m" );
+		LOG(log_error, logtype_papd, "malloc: %s", strerror(errno) );
 		exit (1);
 	}
 
@@ -523,7 +733,7 @@ to_ascii ( char  *inptr, char **outptr)
  * Returns: -1 on failure, length of name on success; outpr contains name in MacRoman
  */
 
-static int convert_to_mac_name ( char * encoding, char * inptr, char * outptr, size_t outlen)
+static int convert_to_mac_name ( const char * encoding, char * inptr, char * outptr, size_t outlen)
 {
 	char   	*outbuf;
 	char	*soptr;
@@ -533,7 +743,7 @@ static int convert_to_mac_name ( char * encoding, char * inptr, char * outptr, s
 
 	/* Change the encoding */
 	if ((charset_t)-1 != (chCups = add_charset(encoding))) {
-		name_len = convert_string_allocate( chCups, CH_MAC, inptr, strlen(inptr), &outbuf);
+		name_len = convert_string_allocate( chCups, CH_MAC, inptr, -1, &outbuf);
 	}
 
 	if (name_len == 0 || name_len == (size_t)-1) {
