@@ -1,5 +1,5 @@
 /*
- * $Id: queries.c,v 1.16.2.1.2.1.2.1 2005/02/06 10:16:02 didg Exp $
+ * $Id: queries.c,v 1.22 2009-10-13 22:55:37 didg Exp $
  *
  * Copyright (c) 1990,1994 Regents of The University of Michigan.
  * All Rights Reserved.  See COPYRIGHT.
@@ -43,9 +43,11 @@ int gq_pagecost( struct papfile * );
 int gq_true( struct papfile * );
 int gq_rbispoolerid( struct papfile * );
 int gq_rbiuamlist( struct papfile * );
+int gq_product( struct papfile * );
 
 int cq_query( struct papfile *, struct papfile * );
 void cq_font_answer( char *, char *, struct papfile * );
+int cq_fontlist(struct papfile *, struct papfile * );
 int cq_font( struct papfile *, struct papfile * );
 int cq_feature( struct papfile *, struct papfile * );
 int cq_printer( struct papfile *, struct papfile * );
@@ -54,11 +56,10 @@ int cq_rmjob( struct papfile *, struct papfile * );
 int cq_listq( struct papfile *, struct papfile * );
 int cq_rbilogin( struct papfile *, struct papfile * );
 #endif  /* HAVE_CUPS */
+int cq_end( struct papfile *, struct papfile * );
 
 
-
-int cq_default( in, out )
-    struct papfile	*in, *out;
+int cq_default( struct papfile *in, struct papfile *out)
 {
     char		*start, *stop, *p;
     int			linelength, crlflength;
@@ -71,6 +72,9 @@ int cq_default( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	stop = start+linelength;
@@ -91,9 +95,11 @@ int cq_default( in, out )
 			break;
 		    }
 		}
-		p++;
-		while ( *p == ' ' ) {
+		if (p < stop) {
 		    p++;
+		    while ( *p == ' ' ) {
+		        p++;
+                    }
 		}
 
 		append( out, p, stop - p + crlflength );
@@ -113,8 +119,7 @@ char	*LoginFailed = "LoginFailed\n";
 
 #define h2b(x)	(isdigit((x))?(x)-'0':(isupper((x))?(x)-'A':(x)-'a')+10)
 
-int cq_k4login( in, out )
-    struct papfile	*in, *out;
+int cq_k4login( struct papfile *in, struct papfile *out)
 {
     char		*start, *p;
     int			linelength, crlflength;
@@ -130,6 +135,9 @@ int cq_k4login( in, out )
 
     case -1 :
 	return( CH_MORE );
+
+    case -2 :
+        return( CH_ERROR );
     }
 
     p = start + strlen( comment->c_begin );
@@ -139,6 +147,7 @@ int cq_k4login( in, out )
 
     bzero( &tkt, sizeof( tkt ));
     stop = start+linelength;
+    /* FIXME */
     for ( i = 0, t = tkt.dat; p < stop; p += 2, t++, i++ ) {
 	*t = ( h2b( (unsigned char)*p ) << 4 ) +
 		h2b( (unsigned char)*( p + 1 ));
@@ -166,8 +175,7 @@ int cq_k4login( in, out )
 
 char	*uameth = "UMICHKerberosIV\n*\n";
 
-int cq_uameth( in, out )
-    struct papfile	*in, *out;
+int cq_uameth( struct papfile *in, struct papfile *out)
 {
     char		*start;
     int			linelength, crlflength;
@@ -180,6 +188,9 @@ int cq_uameth( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	if ( comgetflags() == 0 ) {	/* start */
@@ -204,8 +215,7 @@ int cq_uameth( in, out )
 }
 #endif /* KRB */
 
-int gq_true( out )
-    struct papfile	*out;
+int gq_true( struct papfile *out)
 {
     if ( printer->p_flags & P_SPOOLED ) {
 	append( out, "true\n", 5 );
@@ -215,8 +225,7 @@ int gq_true( out )
     }
 }
 
-int gq_pagecost( out )
-    struct papfile	*out;
+int gq_pagecost( struct papfile *out)
 {
     char		cost[ 60 ];
 
@@ -238,8 +247,7 @@ int gq_pagecost( out )
 }
 
 #ifdef ABS_PRINT
-int gq_balance( out )
-    struct papfile	*out;
+int gq_balance( struct papfile *out)
 {
     char		balance[ 60 ];
 
@@ -259,8 +267,7 @@ int gq_balance( out )
 
 static const char *spoolerid = "(PAPD Spooler) 1.0 (" VERSION ")\n";
 
-int gq_rbispoolerid( out )
-    struct papfile	*out;
+int gq_rbispoolerid( struct papfile *out)
 {
     append( out, spoolerid, strlen( spoolerid ));
     return(0);
@@ -274,8 +281,7 @@ int gq_rbispoolerid( out )
 
 static const char *nouams = "*\n";
 
-int gq_rbiuamlist( out )
-    struct papfile      *out;
+int gq_rbiuamlist( struct papfile *out)
 {
     char uamnames[128] = "\0";
 
@@ -293,6 +299,15 @@ int gq_rbiuamlist( out )
     }
 }
 
+int gq_product( struct papfile *out)
+{
+    struct ppd_feature	*pdprod;
+    pdprod = ppd_feature( "*Product\n", strlen( "*Product\n" ));
+    append( out, pdprod->pd_value, strlen( pdprod->pd_value ));
+    append( out, "\r", 1 );
+    return(0);
+}
+
 
 struct genquery {
     char	*gq_name;
@@ -304,14 +319,14 @@ struct genquery {
 #endif /* notdef */
     { "RBISpoolerID",	gq_rbispoolerid },
     { "RBIUAMListQuery", gq_rbiuamlist },
+    { "Product", gq_product },
     { "ADOIsBinaryOK?", gq_true },
     { "UMICHListQueue", gq_true },
     { "UMICHDeleteJob", gq_true },
     { NULL, NULL },
 };
 
-int cq_query( in, out )
-    struct papfile	*in, *out;
+int cq_query( struct papfile *in, struct papfile *out)
 {
     char		*start, *stop, *p, *q;
     int			linelength, crlflength;
@@ -326,6 +341,9 @@ int cq_query( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	stop = start+linelength;
@@ -339,7 +357,7 @@ int cq_query( in, out )
 		}
 	    }
 
-	    for ( p++; p < stop; p++ ) {
+	    if (p < stop) for ( p++; p < stop; p++ ) {
 		if ( *p != ' ' && *p != '\t' ) {
 		    break;
 		}
@@ -352,7 +370,7 @@ int cq_query( in, out )
 	    }
 
 	    for ( gq = genqueries; gq->gq_name; gq++ ) {
-		if (( strlen( gq->gq_name ) == q - p ) &&
+		if (( strlen( gq->gq_name ) == (size_t)(q - p) ) &&
 			( strncmp( gq->gq_name, p, q - p ) == 0 )) {
 		    break;
 		}
@@ -377,23 +395,23 @@ int cq_query( in, out )
     }
 }
 
-void cq_font_answer( start, stop, out )
-    char		*start, *stop;
-    struct papfile	*out;
+void cq_font_answer( char *start, char *stop, struct papfile *out)
 {
     char		*p, *q, buf[ 256 ];
     struct ppd_font	*pfo;
 
     p = start;
     while ( p < stop ) {
+        unsigned int count = 0;
 	while (( *p == ' ' || *p == '\t' ) && p < stop ) {
 	    p++;
 	}
 
 	q = buf;
 	while ( *p != ' ' && *p != '\t' &&
-		*p != '\n' && *p != '\r' && p < stop ) {
+		*p != '\n' && *p != '\r' && p < stop && count < sizeof(buf)) {
 	    *q++ = *p++;
+	    count++;
 	}
 
 	if ( q != buf ) {
@@ -414,8 +432,77 @@ void cq_font_answer( start, stop, out )
     return;
 }
 
-int cq_font( in, out )
-    struct papfile	*in, *out;
+int cq_fontlist(struct papfile *in, struct papfile *out)
+{
+    char		*start;
+    int			linelength, crlflength;
+    struct papd_comment	*comment = compeek();
+    for (;;){
+        switch ( markline( in, &start, &linelength, &crlflength )) {
+        case 0 :
+	return( 0 );
+
+        case -1 :
+	return( CH_MORE );
+
+        case -2 :
+        return( CH_ERROR );
+        }
+	if ( comgetflags() == 0 ) {
+	    comsetflags( 1 );
+	    /* Send the default 35 PostScript fonts since some cups-filters PPDs lack any font entries. */
+	    append( out,"AvantGarde-Book\n", 16 );
+	    append( out,"AvantGarde-BookOblique\n", 23 );
+	    append( out,"AvantGarde-Demi\n", 16 );
+	    append( out,"AvantGarde-DemiOblique\n", 23 );
+	    append( out,"Bookman-Demi\n", 13 );
+	    append( out,"Bookman-DemiItalic\n", 19 );
+	    append( out,"Bookman-Light\n", 14 );
+	    append( out,"Bookman-LightItalic\n", 20 );
+	    append( out,"Courier\n", 8 );
+	    append( out,"Courier-Bold\n", 13 );
+	    append( out,"Courier-BoldOblique\n", 20 );
+	    append( out,"Courier-Oblique\n", 16 );
+	    append( out,"Helvetica\n", 10 );
+	    append( out,"Helvetica-Bold\n", 15 );
+	    append( out,"Helvetica-BoldOblique\n", 22 );
+	    append( out,"Helvetica-Narrow\n", 17 );
+	    append( out,"Helvetica-Narrow-Bold\n", 22 );
+	    append( out,"Helvetica-Narrow-BoldOblique\n", 29 );
+	    append( out,"Helvetica-Narrow-Oblique\n", 25 );
+	    append( out,"Helvetica-Oblique\n", 18 );
+	    append( out,"NewCenturySchlbk-Bold\n", 22 );
+	    append( out,"NewCenturySchlbk-BoldItalic\n", 28 );
+	    append( out,"NewCenturySchlbk-Italic\n", 24 );
+	    append( out,"NewCenturySchlbk-Roman\n", 23 );
+	    append( out,"Palatino-Bold\n", 14 );
+	    append( out,"Palatino-BoldItalic\n", 20 );
+	    append( out,"Palatino-Italic\n", 16 );
+	    append( out,"Palatino-Roman\n", 15 );
+	    append( out,"Symbol\n", 7 );
+	    append( out,"Times-Bold\n", 11 );
+	    append( out,"Times-BoldItalic\n", 17 );
+	    append( out,"Times-Italic\n", 13 );
+	    append( out,"Times-Roman\n", 12 );
+	    append( out,"ZapfChancery-MediumItalic\n", 26 );
+	    append( out,"ZapfDingbats\n", 13 );
+	    append( out,"*\n", 2 ); /* Terminate the list */
+	    out->pf_state |= PF_FONT_QUERY;
+        }
+        else {
+	    if ( comcmp( start, start+linelength, comment->c_end, 0 ) == 0 ) {
+		compop();
+		CONSUME( in, linelength + crlflength );
+		return( CH_DONE );
+	    }
+
+        }
+    CONSUME( in, linelength + crlflength );
+    }
+}
+
+
+int cq_font(struct papfile *in, struct papfile *out)
 {
     char		*start, *stop, *p;
     int			linelength, crlflength;
@@ -428,6 +515,9 @@ int cq_font( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	stop = start + linelength;
@@ -440,7 +530,8 @@ int cq_font( in, out )
 		    break;
 		}
 	    }
-	    p++;
+	    if (p < stop)
+	        p++;
 
 	    cq_font_answer( p, stop, out );
 	} else {
@@ -453,7 +544,8 @@ int cq_font( in, out )
 			break;
 		    }
 		}
-		p++;
+		if (p < stop)
+		    p++;
 
 		cq_font_answer( p, stop, out );
 	    } else {
@@ -471,8 +563,7 @@ int cq_font( in, out )
     }
 }
 
-int cq_feature( in, out )
-    struct papfile	*in, *out;
+int cq_feature( struct papfile *in, struct papfile *out)
 {
     char		*start, *stop, *p;
     int			linelength, crlflength;
@@ -486,6 +577,9 @@ int cq_feature( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	stop = start + linelength;
@@ -499,11 +593,19 @@ int cq_feature( in, out )
 		    break;
 		}
 	    }
-	    p++;
+	    if (p < stop)
+	        p++;
 	    while ( *p == ' ' ) {
 		p++;
 	    }
 
+	    /* handle the '?' at the beginning of LW8's Resolution Query */
+	    if (*p == '*' && *(p + 1) == '?' && p < stop)
+	    {
+	        p++;
+	        *p = '*';
+	    }
+
 	    if (( pfe = ppd_feature( p, stop - p )) == NULL ) {
 		if ( comswitch( queries, cq_default ) < 0 ) {
 		    LOG(log_error, logtype_papd, "cq_feature: can't find default!" );
@@ -529,8 +631,7 @@ int cq_feature( in, out )
 static const char	*psver = "*PSVersion\n";
 static const char	*prod = "*Product\n";
 
-int cq_printer( in, out )
-    struct papfile	*in, *out;
+int cq_printer(struct papfile *in, struct papfile *out)
 {
     char		*start, *p;
     int			linelength, crlflength;
@@ -544,6 +645,9 @@ int cq_printer( in, out )
 
 	case -1 :
 	    return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
 	}
 
 	if ( comgetflags() == 0 ) {
@@ -607,8 +711,7 @@ int cq_printer( in, out )
 static const char	*rmjobfailed = "Failed\n";
 static const char	*rmjobok = "Ok\n";
 
-int cq_rmjob( in, out )
-    struct papfile	*in, *out;
+int cq_rmjob( struct papfile *in, struct papfile *out)
 {
     char		*start, *stop, *p;
     int			linelength, crlflength;
@@ -620,6 +723,9 @@ int cq_rmjob( in, out )
 
     case -1 :
 	return( CH_MORE );
+
+    case -2 :
+        return( CH_ERROR );
     }
 
     stop = start + linelength;
@@ -648,8 +754,7 @@ int cq_rmjob( in, out )
     return( CH_DONE );
 }
 
-int cq_listq( in, out )
-    struct papfile	*in, *out;
+int cq_listq( struct papfile *in, struct papfile *out)
 {
     char		*start;
     int			linelength, crlflength;
@@ -660,6 +765,9 @@ int cq_listq( in, out )
 
     case -1 :
 	return( CH_MORE );
+
+    case -2 :
+        return( CH_ERROR );
     }
 
     if ( lp_queue( out )) {
@@ -685,14 +793,13 @@ SecurityViolation: Unknown user, incorrect password or log on is \
 disabled ]%%\r%%[Flushing: rest of job (to end-of-file) will be \
 ignored ]%%\r";
 
-int cq_rbilogin( in, out )
-    struct papfile      *in, *out;
+int cq_rbilogin( struct papfile *in, struct papfile *out)
 {
     char        	*start, *stop, *p, *begin;
     int			linelength, crlflength;
     char        	username[UAM_USERNAMELEN + 1] = "\0";
     struct papd_comment	*comment = compeek();
-    char		uamtype[20] = "\0";
+    char		uamtype[20];
 
     for (;;) {
         switch ( markline( in, &start, &linelength, &crlflength )) {
@@ -701,6 +808,9 @@ int cq_rbilogin( in, out )
 
         case -1 :
             return( CH_MORE );
+
+        case -2 :
+            return( CH_ERROR );
         }
 
 	stop = start + linelength;
@@ -709,13 +819,16 @@ int cq_rbilogin( in, out )
 	    begin = start + strlen(comment->c_begin);
 	    p = begin;
 
-	    while (*p != ' ') {
+	    while (*p != ' ' && p < stop) {
 		p++;
 	    }
 
-	    strncat(uamtype, begin, p - begin);
+	    memset(uamtype, 0, sizeof(uamtype));
+	    if ((size_t)(p -begin) <= sizeof(uamtype) -1) {
+	        strncpy(uamtype, begin, p - begin);
+            }
 
-	    if ((papd_uam = auth_uamfind(UAM_SERVER_PRINTAUTH,
+	    if ( !*uamtype || (papd_uam = auth_uamfind(UAM_SERVER_PRINTAUTH,
 				uamtype, strlen(uamtype))) == NULL) {
 		LOG(log_info, logtype_papd, "Could not find uam: %s", uamtype);
 		append(out, rbiloginbad, strlen(rbiloginbad));
@@ -742,6 +855,26 @@ int cq_rbilogin( in, out )
     }
 }
 
+int cq_end( struct papfile *in, struct papfile *out)
+{
+    char                *start;
+    int                 linelength, crlflength;
+
+    switch ( markline( in, &start, &linelength, &crlflength )) {
+    case 0 :
+        return( 0 );
+
+    case -1 :
+        return( CH_MORE );
+
+    case -2 :
+        return( CH_ERROR );
+    }
+    in->pf_state |= PF_QUERY;
+    compop();
+    CONSUME( in, linelength + crlflength );
+    return ( CH_DONE );
+}
 
 /*
  * All queries start with %%?Begin and end with %%?End.  Note that the
@@ -760,7 +893,9 @@ struct papd_comment	queries[] = {
     { "%%?BeginQuery",		"%%?EndQuery",		cq_query,	0 },
     { "%%?BeginFeatureQuery",	"%%?EndFeatureQuery",	cq_feature,	0 },
     { "%%?BeginFontQuery",	"%%?EndFontQuery",	cq_font,	0 },
+    { "%%?BeginFontListQuery",  "%%?EndFontListQuery",  cq_fontlist,    0 },
     { "%%?BeginPrinterQuery",	"%%?EndPrinterQuery",	cq_printer,C_FULL },
     { "%%?Begin",		"%%?End",		cq_default,	0 },
+    { "%%EOF",			NULL,			cq_end,		0 },
     { NULL,			NULL,			NULL,		0 },
 };
